using System;
using System.Linq;
using ELTE.TicTacToeGame.Model;
using ELTE.TicTacToeGame.Persistence;
using Moq;
using Xunit;

namespace ELTE.TicTacToeGame.Test.xUnit
{
    public class TicTacToeModelTest
    {
        private Mock<IPersistence> _mock = null!; // az adatelrs mock-ja
        private TicTacToeModel _model = null!;
        
        public TicTacToeModelTest() // teszt inicializlsa
        {
            _mock = new Mock<IPersistence>();
            _mock.Setup(mock => mock.Load(It.IsAny<String>())).Returns(() => Enumerable.Repeat(Player.NoPlayer, 9).ToArray());
            // a mock a Load mveletben brmilyen paramterre egy res tblnak a tmbjt fogja visszaadni
            // a Save mveletet nem kell konfigurlnunk

            _model = new TicTacToeModel(_mock.Object); // pldnyostjuk a modellt a mock objektummal
        }

        [Fact]
        public void TicTacToeConstructorTest() // egysgteszt mvelet
        {
            _model.NewGame();
            Assert.Equal(3, _model.TableSize);
            for (Int32 i = 0; i < 3; i++)
                for (Int32 j = 0; j < 3; j++)
                    Assert.Equal(Player.NoPlayer, _model[i, j]); // valamennyi mez res
        }

        [Fact]
        public void TicTacToeStepGameTest()
        {
            _model.NewGame();
            _model.StepGame(1, 1);

            Assert.Equal(Player.PlayerX, _model[1, 1]); // tvltott-e az rtk X-re

            for (Int32 i = 0; i < 3; i++)
                for (Int32 j = 0; j < 3; j++)
                {
                    if (i != 1 || j != 1)
                        Assert.Equal(Player.NoPlayer, _model[i, j]); // valamennyi tovbbi mez res
                }

            _model.StepGame(0, 1);

            Assert.Equal(Player.PlayerO, _model[0, 1]); // tvltott-e az rtk O-ra

            Assert.Throws<InvalidOperationException>(() =>
            {
                _model.StepGame(0, 1);
            }); // xUnitban nincs Assert.Fail()

            Assert.Equal(Player.PlayerO, _model[0, 1]); // tovbb nem szabad, hogy az rtk megvltozzon
        }

        [Fact]
        public void TicTacToeStepNumberTest()
        {
            _model.NewGame();
            Assert.Equal(0, _model.StepNumber);

            Int32 k = 0;
            for (Int32 i = 0; i < 3; i++)
                for (Int32 j = 0; j < 2; j++)
                {
                    _model.StepGame(i, j);
                    k++;
                    Assert.Equal(k, _model.StepNumber); // lpsszm megfelel-e
                }
        }

        [Theory]
        [InlineData(0, 0)]
        [InlineData(0, 1)]
        [InlineData(1, 1)]
        [InlineData(2, 0)]
        [InlineData(2, 2)]
        public void TicTacToeIndexerValidTest(int x, int y)
        {
            _model.NewGame();
            Assert.Equal(Player.NoPlayer, _model[x, y]);
            _model.StepGame(x, y);
            Assert.Equal(Player.PlayerX, _model[x, y]);
        }

        [Theory]
        [InlineData(-1, 0)]
        [InlineData(3, 1)]
        [InlineData(1, -1)]
        [InlineData(2, 3)]
        public void TicTacToeIndexerInvalidTest(int x, int y)
        {
            _model.NewGame();

            Assert.Throws<ArgumentException>(() =>
            {
                Player value = _model[x, y];
            });
        }

        [Fact]
        public void TicTacToeGameWonTest()
        {
            bool eventRaised = false;
            _model.GameWon += delegate (object? sender, GameWonEventArgs e)
            {
                eventRaised = true;
                Assert.True(e.Player == Player.PlayerX); // a megfelel jtkos gyztt-e
            };

            _model.NewGame();
            _model.StepGame(0, 0);
            _model.StepGame(0, 1);
            _model.StepGame(1, 0);
            _model.StepGame(1, 1);
            _model.StepGame(2, 0); // hagyjuk az X jtkost nyerni

            Assert.True(eventRaised); // kivltottuk-e az esemnyt
        }

        [Fact]
        public void TicTacToeGameLoadTest()
        {
            // kezdnk egy j jtkot s vgrehajtunk pr lpst
            _model.NewGame();
            _model.StepGame(0, 0);
            _model.StepGame(0, 1);
            _model.StepGame(1, 0);
            _model.StepGame(1, 1);

            // majd betltnk egy jtkot
            _model.LoadGame(String.Empty);

            Int32 stepNumber = 0;
            for (Int32 i = 0; i < 3; i++)
                for (Int32 j = 0; j < 3; j++)
                {
                    Assert.Equal(Player.NoPlayer, _model[i, j]); // ellenrizzk, valamennyi mez res
                    stepNumber += (_model[i, j] != Player.NoPlayer) ? 1 : 0;
                }

            // ellenrizzk a lpsszmot, s a kvetkez  jtkost
            Assert.Equal(stepNumber, _model.StepNumber);
            Assert.Equal(stepNumber % 2 == 0 ? Player.PlayerX : Player.PlayerO, _model.CurrentPlayer);

            // ellenrizzk, hogy meghvtk-e a Load mveletet a megadott paramterrel
            _mock.Verify(dataAccess => dataAccess.Load(String.Empty), Times.Once());
        }

        [Fact]
        public void TicTacToeGameSaveTest()
        {
            // kimentjk a jtkllapotot
            Player currentPlayer = _model.CurrentPlayer;
            Int32 stepNumber = _model.StepNumber;
            Player[] values = new Player[9];

            for (Int32 i = 0; i < 3; i++)
                for (Int32 j = 0; j < 3; j++)
                {
                    values[i * 3 + j] = _model[i, j];
                }

            // vgrehajtjuk a mentst
            _model.SaveGame(String.Empty);

            // ellenrizzk, hogy a ments utn nem vltozott a jtkllapot
            Assert.Equal(currentPlayer, _model.CurrentPlayer);
            Assert.Equal(stepNumber, _model.StepNumber);

            for (Int32 i = 0; i < 3; i++)
                for (Int32 j = 0; j < 3; j++)
                {
                    Assert.Equal(values[i * 3 + j], _model[i, j]);
                }

            // ellenrizzk, hogy meghvtk-e a Save mveletet
            _mock.Verify(mock => mock.Save(String.Empty, It.IsAny<Player[]>()), Times.Once());
        }
    }
}
